"""
Base interface for any kind of encoder.

This is basically any generic function x -> y.

Note that in practice, when designing some model,
this interface is even not needed,
because you only care about the final encoded vectors,
and not how you got there.
Automatic differentiation will automatically
also train the encoder.
So, for most purpose, e.g. for a decoder (see :mod:`..decoder.base`),
you only care about some encoded vector of type :class:`Tensor`.
"""

from __future__ import annotations
from typing import Tuple, Union
from abc import ABC
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


class IEncoder(rf.Module, ABC):
    """
    Generic encoder interface

    The encoder is a function x -> y.
    The input can potentially be sparse or dense.
    The output is dense with feature dim `out_dim`.
    """

    out_dim: Dim

    def __call__(self, source: Tensor) -> Tensor:
        """
        Encode the input
        """
        raise NotImplementedError


class ISeqFramewiseEncoder(rf.Module, ABC):
    """
    This specializes IEncoder that it operates on a sequence.
    The output sequence length here is the same as the input.
    """

    out_dim: Dim

    def __call__(self, source: Tensor, *, spatial_dim: Dim) -> Tensor:
        raise NotImplementedError


class ISeqDownsamplingEncoder(rf.Module, ABC):
    """
    This is more specific than IEncoder in that it operates on a sequence.
    The output sequence length here is shorter than the input.

    This is a common scenario for speech recognition
    where the input might be on 10ms/frame
    and the output might cover 30ms/frame or 60ms/frame or so.
    """

    out_dim: Dim
    # In most cases (pooling, conv), the output sequence length will bei ceildiv(input_seq_len, factor)
    # and factor is an integer.
    # However, this is not a hard condition.
    # The downsampling factor only describes the linear factor in the limit.
    downsample_factor: Union[int, float]

    def __call__(self, source: Tensor, *, in_spatial_dim: Dim) -> Tuple[Tensor, Dim]:
        raise NotImplementedError



"""
container functions
"""

from __future__ import annotations
import returnn.frontend as rf
from returnn.tensor import Tensor
from typing import Optional, TypeVar, Generic, Iterable, Iterator, Union, Tuple, Dict, Callable


__all__ = ["ModuleList", "Sequential", "sequential", "ParameterList"]


_UnaryFuncT = Callable[[Tensor], Tensor]
_ModT = Union[rf.Module, _UnaryFuncT]
__ModT = TypeVar("__ModT", bound=rf.Module)


class ModuleList(rf.Module, Generic[__ModT]):
    """
    Module list, getting passed an Iterable of Modules and creates a list of Modules in that order
    """

    def __init__(self, *modules: Union[__ModT, Iterable[__ModT], Dict[str, __ModT], ModuleList]):
        super().__init__()
        if len(modules) == 1 and isinstance(modules[0], dict):
            for i, (key, module) in enumerate(modules[0].items()):
                if _is_int_str(key):
                    key = str(i)
                setattr(self, key, _convert_to_module(module))
        elif len(modules) == 1 and isinstance(modules[0], ModuleList):
            for key, module in modules[0]._get_modules().items():
                setattr(self, key, _convert_to_module(module))
        elif len(modules) == 1 and _is_iterable(modules[0]):
            for idx, module in enumerate(modules[0]):
                setattr(self, str(idx), _convert_to_module(module))
        else:
            for idx, module in enumerate(modules):
                setattr(self, str(idx), _convert_to_module(module))

    def _get_modules(self) -> Dict[str, __ModT]:
        return {key: value for (key, value) in vars(self).items() if isinstance(value, rf.Module)}

    def append(self, module: __ModT) -> ModuleList[__ModT]:
        """
        appends one module to the list
        """
        setattr(self, str(len(self)), _convert_to_module(module))
        return self

    def extend(self, modules: Iterable[__ModT]) -> ModuleList[__ModT]:
        """
        appends multiple modules to the list
        """
        for module in modules:
            self.append(module)
        return self

    def __len__(self) -> int:
        return len(self._get_modules())

    def __iter__(self) -> Iterator[__ModT]:
        return iter(self._get_modules().values())

    def items(self) -> Iterable[Tuple[str, __ModT]]:
        """module items"""
        return self._get_modules().items()

    def __getitem__(self, idx: Union[slice, int]) -> Union[ModuleList[__ModT], __ModT]:
        if isinstance(idx, slice):
            return self.__class__(dict(list(self._get_modules().items())[idx]))
        else:
            return list(self._get_modules().values())[idx]

    def __setitem__(self, idx: Union[slice, int], module: Union[__ModT, Iterable[__ModT]]) -> None:
        key = list(self._get_modules().keys())[idx]
        if isinstance(idx, slice):
            assert not idx.step or idx.step == 1  # not supported
            mod_items = list(self._get_modules().items())
            if idx.stop is not None:
                remaining = mod_items[idx.stop :]
            else:
                remaining = []
            # Delete also remaining, and then re-add them later, such that indices are correct.
            for k, _ in mod_items[idx.start :]:
                delattr(self, k)
            i = idx.start or 0
            for mod_ in module:
                assert not hasattr(self, str(i))
                setattr(self, str(i), _convert_to_module(mod_))
                i += 1
            for k, v in remaining:
                if _is_int_str(k):
                    k = str(i)
                assert not hasattr(self, k)
                setattr(self, k, v)
                i += 1
        else:
            setattr(self, key, _convert_to_module(module))

    def __delitem__(self, key: Union[slice, int]):
        if isinstance(key, slice):
            self[key] = []
        else:
            self[key : key + 1] = []

    __call__ = rf.Module.__call__  # stays abstract


class Sequential(ModuleList):
    """
    Sequential Module, takes callable of Modules which are then executed in sequence
    """

    def __call__(self, inp, *, collected_outputs: Optional[Dict[str, Tensor]] = None, **kwargs) -> Tensor:
        """
        Forward
        """
        for name, module in self.items():
            inp = module(inp, **kwargs)
            if collected_outputs is not None:
                collected_outputs[name] = inp
        return inp


def sequential(source: Tensor, *modules) -> Tensor:
    """
    Wraps ``Sequential(*modules)(source)``
    """
    return Sequential(*modules)(source)


def _convert_to_module(obj: _ModT) -> rf.Module:
    if isinstance(obj, rf.Module):
        return obj
    elif callable(obj):
        return rf.Functional(obj)
    else:
        raise TypeError(f"Expected rf.Module or callable, did not expect {obj!r} ({type(obj)})")


def _is_iterable(obj) -> bool:
    # No good generic way, so do this ugly hack.
    # https://stackoverflow.com/questions/1952464/in-python-how-do-i-determine-if-an-object-is-iterable
    try:
        iter(obj)
        return True
    except TypeError:
        return False


class ParameterList(rf.Module):
    """
    Parameter list, getting passed an Iterable of Parameters and creates a list of Parameters in that order
    """

    def __init__(
        self, *parameters: Union[rf.Parameter, Iterable[rf.Parameter], Dict[str, rf.Parameter], ParameterList]
    ):
        super().__init__()
        if len(parameters) == 1 and isinstance(parameters[0], dict):
            for i, (key, parameter) in enumerate(parameters[0].items()):
                if _is_int_str(key):
                    key = str(i)
                setattr(self, key, parameter)
        elif len(parameters) == 1 and isinstance(parameters[0], ParameterList):
            for key, parameter in parameters[0]._get_parameters().items():
                setattr(self, key, parameter)
        elif len(parameters) == 1 and _is_iterable(parameters[0]):
            for idx, parameter in enumerate(parameters[0]):
                setattr(self, str(idx), parameter)
        else:
            for idx, parameter in enumerate(parameters):
                setattr(self, str(idx), parameter)

    def _get_parameters(self) -> Dict[str, rf.Parameter]:
        return {key: value for (key, value) in vars(self).items() if isinstance(value, rf.Parameter)}

    def append(self, parameter: rf.Parameter) -> ParameterList:
        """
        appends one Parameter to the list
        """
        setattr(self, str(len(self)), parameter)
        return self

    def extend(self, parameters: Iterable[rf.Parameter]) -> ParameterList:
        """
        appends multiple Parameters to the list
        """
        for parameter in parameters:
            self.append(parameter)
        return self

    def __len__(self) -> int:
        return len(self._get_parameters())

    def __iter__(self) -> Iterator[rf.Parameter]:
        return iter(self._get_parameters().values())

    def __getitem__(self, idx) -> Union[ParameterList, rf.Parameter]:
        if isinstance(idx, slice):
            return self.__class__(dict(list(self._get_parameters().items())[idx]))
        else:
            return list(self._get_parameters().values())[idx]

    def __setitem__(self, idx: int, parameter: rf.Parameter) -> None:
        key = list(self._get_parameters().keys())[idx]
        return setattr(self, key, rf.Parameter)

    __call__ = rf.Module.__call__  # stays abstract


def _is_int_str(s: str) -> bool:
    try:
        int(s)
        return True
    except ValueError:
        return False



"""
Watch memory usage over time.
"""

from __future__ import annotations
from typing import Dict
import time
from collections import defaultdict
from threading import Thread
import psutil  # noqa


def watch_memory():
    """
    Start thread which watches memory usage over time of the current process and all its children over time.
    """
    global _watch_memory_thread
    if _watch_memory_thread:
        return
    _watch_memory_thread = Thread(target=_watch_memory_thread_main, name="watch_memory", daemon=True)
    _watch_memory_thread.start()


_watch_memory_thread = None


def _watch_memory_thread_main():
    prefix = "MEMORY:"
    cur_proc = psutil.Process()
    procs = []
    mem_per_pid = {}

    while True:
        change = False
        procs_ = [cur_proc] + cur_proc.children(recursive=True)
        for p in procs:
            if p not in procs_:
                print(prefix, f"proc {_format_proc(p)} exited, old:", _format_mem_info(mem_per_pid[p.pid]))
                mem_per_pid.pop(p.pid, None)
                change = True
        procs = procs_

        for p in list(procs):
            old_mem_info = mem_per_pid.get(p.pid, None)
            try:
                mem_info = get_mem_info(p)
            except psutil.NoSuchProcess:  # race condition, can happen
                if old_mem_info:
                    print(prefix, f"proc {_format_proc(p)} exited, old:", _format_mem_info(old_mem_info))
                    mem_per_pid.pop(p.pid, None)
                    change = True
                procs.remove(p)
                continue
            proc_prefix = "main" if p == cur_proc else "sub"
            if not old_mem_info:
                print(prefix, f"{proc_prefix} proc {_format_proc(p)} initial:", _format_mem_info(mem_info))
                mem_per_pid[p.pid] = mem_info
                change = True
            elif mem_info["rss"] > old_mem_info["rss"] and _format_mem_size(old_mem_info["rss"]) != _format_mem_size(
                mem_info["rss"]
            ):
                print(prefix, f"{proc_prefix} proc {_format_proc(p)} increased RSS:", _format_mem_info(mem_info))
                # keep old info otherwise, such that the update check works
                mem_per_pid[p.pid] = mem_info
                change = True

        if change:
            res = {"pss": 0, "uss": 0}
            for mem_info in mem_per_pid.values():
                for k in res.keys():
                    res[k] += mem_info[k]
            print(prefix, f"total ({len(mem_per_pid)} procs):", _format_mem_info(res))

        time.sleep(5)


def _format_proc(proc: psutil.Process) -> str:
    try:
        proc_name = proc.name()
    except psutil.NoSuchProcess:  # race condition
        proc_name = getattr(proc, "_name", None)
        if not proc_name:
            proc_name = "<unknown-dead>"
    if not proc_name:
        proc_name = "<noname>"
    return "%s(%s)" % (proc_name, proc.pid)


def _format_mem_info(info: Dict[str, int]) -> str:
    return " ".join("%s=%s" % (k, _format_mem_size(v)) for (k, v) in info.items())


def _format_mem_size(c: int) -> str:
    if c < 1024:
        return "%iB" % c
    units = "KMG"
    i = 0
    while i < len(units) - 1:
        if c < 0.8 * 1024 ** (i + 2):
            break
        i += 1
    f = float(c) / (1024 ** (i + 1))
    return "%.1f%sB" % (f, units[i])


def get_mem_info(proc: psutil.Process) -> Dict[str, int]:
    """
    Code from:
    https://ppwwyyxx.com/blog/2022/Demystify-RAM-Usage-in-Multiprocess-DataLoader/
    """
    res = defaultdict(int)
    for mmap in proc.memory_maps():
        res["rss"] += mmap.rss
        res["pss"] += mmap.pss
        res["uss"] += mmap.private_clean + mmap.private_dirty
        res["shared"] += mmap.shared_clean + mmap.shared_dirty
    return res



#!/usr/bin/env python3

"""
Inspect checkpoint
"""

from __future__ import annotations
from typing import Union, Any, Tuple
import argparse
import numpy
import torch


def main():
    """main"""
    # First set PyTorch default print options. This might get overwritten by the args below.
    numpy.set_printoptions(precision=4, linewidth=80)

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("checkpoint")
    arg_parser.add_argument("--key", type=str, default="", help="Name of the tensor or object to inspect")
    arg_parser.add_argument("--all_tensors", action="store_true", help="If True, print the values of all the tensors.")
    arg_parser.add_argument("--stats_only", action="store_true")
    arg_parser.add_argument(
        "--printoptions",
        nargs="*",
        type=parse_numpy_printoption,
        help="Argument for numpy.set_printoptions(), in the form 'k=v'.",
    )
    arg_parser.add_argument("--device", default="cpu")
    arg_parser.add_argument("--mmap", action="store_true")
    args = arg_parser.parse_args()

    state = torch.load(args.checkpoint, map_location=args.device, mmap=args.mmap)
    if args.key:
        assert isinstance(state, dict)
        if args.key not in state and "model" in state:
            state = state["model"]
        obj = state[args.key]
        print(f"{args.key}:")
        print_object(obj, stats_only=args.stats_only)
    else:
        print_object(state, print_all_tensors=args.all_tensors, stats_only=args.stats_only)


def print_object(obj: Any, *, print_all_tensors: bool = False, stats_only: bool = False, prefix: str = ""):
    """print object"""
    if isinstance(obj, (dict, list, tuple)):
        for k, v in obj.items() if isinstance(obj, dict) else enumerate(obj):
            _print_key_value(k, v, print_all_tensors=print_all_tensors, stats_only=stats_only, prefix=prefix)
    elif isinstance(obj, (numpy.ndarray, torch.Tensor)):
        print_tensor(obj, stats_only=stats_only, prefix=prefix)
    else:
        print(f"{prefix}({type(obj)}) {obj}")


def _print_key_value(
    k: Any,
    v: Union[numpy.ndarray, torch.Tensor],
    *,
    print_all_tensors: bool = False,
    stats_only: bool = False,
    prefix: str = "",
):
    if isinstance(v, numpy.ndarray):
        v = torch.tensor(v)
    if isinstance(v, torch.Tensor):
        if v.numel() <= 1 and v.device.type != "meta":
            print(f"{prefix}{k}: {v.dtype} {_format_shape(v.shape)} {_r(v)}")
        else:
            print(f"{prefix}{k}: {v.dtype} {_format_shape(v.shape)}")
            if print_all_tensors:
                print_tensor(v, with_type_and_shape=False, stats_only=stats_only, prefix=prefix + "  ")
    elif isinstance(v, (dict, list, tuple)):
        print(f"{prefix}{k}: ({type(v).__name__})")
        print_object(v, print_all_tensors=print_all_tensors, stats_only=stats_only, prefix=prefix + "  ")
    else:
        print(f"{prefix}{k}: ({type(v).__name__}) {v}")


def print_tensor(
    v: Union[numpy.ndarray, torch.Tensor],
    *,
    prefix: str = "",
    with_type_and_shape: bool = True,
    stats_only: bool = False,
):
    """print tensor"""
    if isinstance(v, numpy.ndarray):
        v = torch.tensor(v)
    assert isinstance(v, torch.Tensor)
    if with_type_and_shape:
        print(f"{prefix}{v.dtype}, {_format_shape(v.shape)}")
    if not stats_only:
        print(v.detach().cpu().numpy())
    n = v.numel()
    if n > 1:
        if v.is_floating_point():
            # See :func:`variable_scalar_summaries_dict`.
            mean = torch.mean(v)
            print(f"{prefix}mean, stddev: {_r(mean)}, {_r(torch.sqrt(torch.mean(torch.square(v - mean))))}")
        v_, _ = v.flatten().sort()
        print(
            f"{prefix}min, p05, p50, p95, max:"
            f" {_r(v_[0])},"
            f" {', '.join(_r(v_[int(n * q)]) for q in (0.05, 0.5, 0.95))},"
            f" {_r(v_[-1])}"
        )


def _format_shape(shape: Tuple[int, ...]) -> str:
    return "[%s]" % ",".join(map(str, shape))


def _r(num: torch.Tensor) -> str:
    return numpy.array2string(num.detach().cpu().numpy())


def parse_numpy_printoption(kv_str):
    """Sets a single numpy printoption from a string of the form 'x=y'.

    See documentation on numpy.set_printoptions() for details about what values
    x and y can take. x can be any option listed there other than 'formatter'.

    Args:
      kv_str: A string of the form 'x=y', such as 'threshold=100000'

    Raises:
      argparse.ArgumentTypeError: If the string couldn't be used to set any
          nump printoption.
    """
    k_v_str = kv_str.split("=", 1)
    if len(k_v_str) != 2 or not k_v_str[0]:
        raise argparse.ArgumentTypeError("'%s' is not in the form k=v." % kv_str)
    k, v_str = k_v_str
    printoptions = numpy.get_printoptions()
    if k not in printoptions:
        raise argparse.ArgumentTypeError("'%s' is not a valid printoption." % k)
    v_type = type(printoptions[k])
    if printoptions[k] is None:
        raise argparse.ArgumentTypeError("Setting '%s' from the command line is not supported." % k)
    try:
        v = v_type(v_str) if v_type is not bool else _to_bool(v_str)
    except ValueError as e:
        raise argparse.ArgumentTypeError(str(e))
    numpy.set_printoptions(**{k: v})


def _to_bool(s: str) -> bool:
    """
    :param s: str to be converted to bool, e.g. "1", "0", "true", "false"
    :return: boolean value, or fallback
    """
    s = s.lower()
    if s in ["1", "true", "yes", "y"]:
        return True
    if s in ["0", "false", "no", "n"]:
        return False
    raise ValueError(f"invalid bool: {s!r}")


if __name__ == "__main__":
    main()
